import torch
import torch.nn.functional as F
from .base_defense import BaseDefense
from .regularizers import r_drop_regularizer
from replay.isotropic_replay import isotropic_replay
from replay.anisotropic_replay import anisotropic_replay


# This AIRDefense class implements
# "Anisotropic & Isotropic Replay" for continual adversarial defense.

class AIRDefense(BaseDefense):
    def __init__(self, student_model, teacher_model, lambda_SD=1.0, lambda_IR=1.0, lambda_AR=1.0, lambda_Reg=1.0,
                 alpha_range=(0.3, 0.7), use_rdrop=True, iso_noise_std=0.01, iso_clamp_min=0.0, iso_clamp_max=1.0,
                 iso_p_flip=0.5, iso_flip_dim=3, iso_p_rotation=0.5, iso_max_rotation=180, iso_p_crop=0.5,
                 iso_p_erase=0.5):
        # We'll call BaseDefense constructor and store all important hyperparams
        super().__init__(student_model, teacher_model)
        self.lambda_SD = lambda_SD
        self.lambda_IR = lambda_IR
        self.lambda_AR = lambda_AR
        self.lambda_Reg = lambda_Reg
        self.alpha_range = alpha_range
        self.use_rdrop = use_rdrop

        self.iso_noise_std = iso_noise_std
        self.iso_clamp_min = iso_clamp_min
        self.iso_clamp_max = iso_clamp_max
        self.iso_p_flip = iso_p_flip
        self.iso_flip_dim = iso_flip_dim
        self.iso_p_rotation = iso_p_rotation
        self.iso_max_rotation = iso_max_rotation
        self.iso_p_crop = iso_p_crop
        self.iso_p_erase = iso_p_erase

    def loss_function(self, x_adv, y, optimizer=None):
        # 1) Compute standard adversarial training loss
        student_logits_new = self.student_model(x_adv)
        loss_AT = F.cross_entropy(student_logits_new, y)

        # 2) Isotropic replay: random augmentation near x_adv
        X_IR, teacher_logits_IR = isotropic_replay(x_adv, teacher_model=self.teacher_model,
            clamp_min=self.iso_clamp_min, clamp_max=self.iso_clamp_max, noise_std=self.iso_noise_std,
            p_flip=self.iso_p_flip, flip_dim=self.iso_flip_dim, p_rotation=self.iso_p_rotation,
            max_rotation=self.iso_max_rotation, p_crop=self.iso_p_crop, p_erase=self.iso_p_erase)
        student_logits_IR = self.student_model(X_IR)
        loss_IR = F.kl_div(F.log_softmax(student_logits_IR, dim=1), F.softmax(teacher_logits_IR, dim=1),
            reduction='batchmean')

        # 3) Anisotropic replay: mix each sample with a shuffled version
        X_AR, teacher_logits_AR = anisotropic_replay(x_adv, teacher_model=self.teacher_model,
            alpha_range=self.alpha_range)
        student_logits_AR = self.student_model(X_AR)
        loss_AR = F.kl_div(F.log_softmax(student_logits_AR, dim=1), F.softmax(teacher_logits_AR, dim=1),
            reduction='batchmean')

        # 4) Optional R-Drop regularization
        loss_Reg = torch.tensor(0.0, device=x_adv.device)
        if self.use_rdrop:
            loss_Reg = r_drop_regularizer(self.student_model, x_adv, y)

        # Combine all terms: adv training, self-distillation (IR+AR), and reg
        loss_AIR = loss_AT + self.lambda_SD * (loss_IR + loss_AR) + self.lambda_Reg * loss_Reg

        return loss_AIR
