import torch
import torch.nn.functional as F


# This r_drop_regularizer function performs multiple forward passes
# with dropout to compute pairwise KL divergences between each pair
# of logits. It doesn't add cross-entropy to the real labels here;
# that might be done elsewhere in the pipeline.

def r_drop_regularizer(model, x, y, num_passes=2):
    # If we only do one pass, there's no pairwise comparison to do.
    if num_passes < 2:
        return torch.tensor(0.0, device=x.device)

    # Put the model in train mode so dropout is active
    # (Though sometimes we want to keep the train/eval mode
    # consistent with the main loop. This is a design choice.)
    model.train()

    logits_list = []
    # We'll collect logits for multiple forward passes
    for _ in range(num_passes):
        logits_list.append(model(x))

    total_kl = 0.0
    count = 0
    # Compare each pair (i,j) of passes using KL-divergence in both directions
    for i in range(num_passes):
        for j in range(i + 1, num_passes):
            kl1 = F.kl_div(F.log_softmax(logits_list[i], dim=1), F.softmax(logits_list[j], dim=1),
                reduction='batchmean')
            kl2 = F.kl_div(F.log_softmax(logits_list[j], dim=1), F.softmax(logits_list[i], dim=1),
                reduction='batchmean')
            # We average kl1 and kl2 so that it’s symmetric
            total_kl += 0.5 * (kl1 + kl2)
            count += 1

    # If no pairs (i.e., num_passes=1), just return zero.
    if count == 0:
        return torch.tensor(0.0, device=x.device)

    # Return the average KL across all pairs
    return total_kl / count
